/* $Id: mon_serial.c,v 1.3 2002/01/13 16:58:57 amura Exp $ */
/*
 *  Copyright (c) 2001,2002    MURAMATSU Atsushi
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * A Simple Monitor Program
 *   - serial I/O handling routines
 */

#include "macdep.h"
#include "monitor_in.h"
#include "monitor.h"

static volatile unsigned char send_buffer[S_SEND_BUFFER_SIZE];
static volatile int send_buf_read_idx;
static volatile int send_buf_write_idx;
static volatile unsigned char send_special_char;
static volatile int send_suspend;

static volatile unsigned char recv_buffer[S_RECV_BUFFER_SIZE];
static volatile int recv_buf_read_idx;
static volatile int recv_buf_write_idx;
static volatile int recv_buf_size;
static volatile int receive_suspend;

static long bit_rate;
static int handshake;

#ifdef __GNUC__
void __recv_routine(void) __attribute__((interrupt_handler));
void __send_routine(void) __attribute__((interrupt_handler));
#ifdef HAVE_SLINE_INTR
void __line_check(void) __attribute__((interrupt_handler));
#endif
#else /* HITACHI SH-C */
#pragma interrupt(__recv_routine)
#pragma interrupt(__send_routine)
#ifdef HAVE_SLINE_INTR
#pragma interrupt(__line_check)
#endif
#endif

#if defined(HAVE_RTSCTS) && !(defined(HAVE_SLINE_INTR)||defined(HAVE_SLINE_AUTO))
#define HAVE_SLINE_HAND
#endif

#if defined(HAVE_RTSCTS) && !defined(HAVE_SLINE_AUTO)
void
__line_check(void)
{
    register int cts;
#ifndef HAVE_SLINE_HAND
    SERIAL_INTR_RESET();
#endif
    if (handshake == S_RTSCTS) {
	cts = GET_SERIAL_CTS();
	if (!cts && send_suspend) {
	    if (send_buf_read_idx != send_buf_write_idx)
		SEND_START();
	    send_suspend = FALSE;
	}
	else if (cts && !send_suspend) {
	    SEND_STOP();
	    send_suspend = TRUE;
	}
    }
}
#endif

void
__send_routine(void)
{
    SERIAL_INTR_RESET();
    if (send_special_char != '\0') {
	SEND(send_special_char);
	send_special_char = '\0';
	return;
    }
    if (send_buf_read_idx == send_buf_write_idx) {
	SEND_END();
	return;
    }
    SEND(send_buffer[send_buf_read_idx]);
    send_buf_read_idx = (send_buf_read_idx + 1) % S_SEND_BUFFER_SIZE;
#if defined(HAVE_RTSCTS) && defined(HAVE_SLINE_HAND)
    __line_check();
#endif
    return;
}

void
__recv_routine(void)
{
    unsigned char c;
    SERIAL_INTR_RESET();
    if (IS_ERROR()) {
	CLEAR_ERROR();
	return;
    }
    RECV(c);
    if (recv_buf_size >= S_RECV_BUFFER_SIZE) {
	/* drop data */
	return;
    }
    if (handshake && recv_buf_size >= S_RECV_BUFFER_WARN && !receive_suspend) {
#ifdef HAVE_RTSCTS
	if (handshake == S_RTSCTS)
	    SET_SERIAL_RTS(OFF);
	else /* if (handshake == S_XONOFF) */
#endif
	{
	    send_special_char = XOFF;
	    SEND_START();
	}
	receive_suspend = TRUE;
    }
#if defined(HAVE_RTSCTS) && defined(HAVE_SLINE_HAND)
    __line_check();
#endif
    if (handshake == S_XONOFF) {
	if (c == XON) {
	    send_suspend = FALSE;
	    if (send_buf_write_idx != send_buf_read_idx)
		SEND_START();
	    return;
	}
	else if (c == XOFF) {
	    send_suspend = TRUE;
	    SEND_END();
	    return;
	}
    }
    recv_buffer[recv_buf_write_idx] = c;
    recv_buf_size++;
    recv_buf_write_idx = (recv_buf_write_idx + 1) % S_RECV_BUFFER_SIZE;
    return;
}

void
serial_init(void)
{
    send_buf_read_idx = send_buf_write_idx = 0;
    send_special_char = '\0';
    send_suspend = FALSE;

    recv_buf_read_idx = recv_buf_write_idx = 0;
    recv_buf_size = 0;
    receive_suspend = FALSE;

    bit_rate  = DEFAULT_BIT_RATE;
    handshake = S_NOFLOW;
    __serial_init();
}

void
serial_mode(long speed, int type)
{
    if (bit_rate == speed && handshake == type)
	return;
    __serial_mode(speed, type);
    bit_rate  = speed;
    handshake = type;
}

void
s_putc(int c)
{
    register int new_write_idx = (send_buf_write_idx + 1) % S_SEND_BUFFER_SIZE;
    while (new_write_idx == send_buf_read_idx);
    send_buffer[send_buf_write_idx] = c;
    send_buf_write_idx = new_write_idx;
    if (!send_suspend)
	SEND_START();
}

void
s_puts(const register char *str)
{
    while (*str) {
	s_putc(*str & 0xFF);
	str++;
    }
}

int
s_getc(void)
{
    int c;
    while (recv_buf_size == 0);
    c = recv_buffer[recv_buf_read_idx] & 0xFF;
    recv_buf_read_idx = (recv_buf_read_idx + 1) % S_RECV_BUFFER_SIZE;
    recv_buf_size--;
    if (handshake && receive_suspend && recv_buf_size < S_RECV_BUFFER_NORM) {
#ifdef HAVE_RTSCTS
	if (handshake == S_RTSCTS)
	    SET_SERIAL_RTS(ON);
	else /* if (handshake == S_XONOFF) */
#endif
	{
	    send_special_char = XON;
	    SEND_START();
	}
	receive_suspend = FALSE;
    }
    return c;
}

int
s_getc_nowait(void)
{
    if (recv_buf_size == 0)
	return -1;
    return s_getc();
}

int
s_gets(register char *str, int strsize)
{
    register int i, c;
    for (i=0; i<strsize-1; i++) {
	c = s_getc();
	*str++ = c;
	if (c == '\n' || c == '\r') {
	    *str = '\0';
	    return i;
	}
    }
    *str = '\0';
    return i+1;
}

int
s_gets_echo(register char *str, int strsize)
{
    register int i, c;
    i = 0;
    while (TRUE) {
	c = s_getc();
	if (c == BACKSPACE) {
	    if (i != 0) {
		str--; i--;
		s_putc(BACKSPACE); s_putc(' '); s_putc(BACKSPACE);
	    }
	    continue;
	}
	if (i == strsize-1)
	    continue;
	i++;
	s_putc(c);
	*str++ = c;
	if (c == '\n' || c == '\r')
	    break;
    }
    *str = '\0';
    return i;
}

int
s_gets_echo_with_go(register char *str, int strsize)
{
    register int i, c;
    i = 0;
    while (TRUE) {
	do {
	    if (CHECK_GO_SWITCH())
		return -1;
	    c = s_getc_nowait();
	} while (c < 0);
	if (c == BACKSPACE) {
	    if (i != 0) {
		str--; i--;
		s_putc(BACKSPACE); s_putc(' '); s_putc(BACKSPACE);
	    }
	    continue;
	}
	if (i == strsize-1)
	    continue;
	i++;
	s_putc(c);
	*str++ = c;
	if (c == '\n' || c == '\r')
	    break;
    }
    *str = '\0';
    return i;
}
